import numpy as np
import cmath as cm

class CustomError(Exception):
    '''
    The class is used to raise custom expections
    '''
    pass

def convert_to_polar(cart):
    '''
    docs: Convert complex number in the cartesian form
          into polar form.
    :inputs:
    cart: Complex number in cartesian number ex. 12+23j
         -> <class 'complex'>
    output: Complex number in polar form (modulus, phase in degrees)
            ex.(12, 90) -> <class 'tuple'>
    '''
    phase = cm.phase(cart) * 180 / cm.pi
    if phase<0:
            phase= phase+360
    return (abs(cart), phase)

def convert_to_cartesian(polar):
    '''
    docs: Convert number from polar form to cartesian complex number.
    :inputs:
    polar: Complex number in polar form (modulus, phase in degrees)
            ex.(12, 90) -> <class 'tuple'>
    output: Complex number in cartesian number ex. 12+23j
         -> <class 'complex'>
    '''
    theta = polar[1] * cm.pi / 180
    return complex(polar[0]*cm.cos(theta), polar[0] * cm.sin(theta))


def convert_math_cart(str):
    '''
    docs: Convert from polar mathematical expression to cartesian.
    :inputs:
    str: mathematical expression for polar number in form of
        modulus@angle(degrees) ex. 12@20 -> 12 at 20 degree angle.
    output: Complex number in cartesian number ex. 12+23j
         -> <class 'complex'>
    '''
    # TODO BUG: accept real numbers without @
    polar = tuple(map(float, str.split(sep='@')))
    return convert_to_cartesian(polar)

convert_math_cart = np.vectorize(convert_math_cart)

def convert_cart_math(complex_num):
    """TODO: inverse of convert_math_cart

    :complex: TODO
    :returns: TODO

    """
    polar = convert_to_polar(complex_num)
    return (str(round(polar[0], 3)))+'@'+(str(round(polar[1], 1)))

convert_cart_math = np.vectorize(convert_cart_math)
def convert_matrix_to_cart(ALPHA_math):
    """
    docs: Convert influence coeffecient matrix ALPHA from mathematical expression
    form to cartesian form.

    :ALPHA_math: list of lists with polar mathmematical expression
            ex . [[90@58, 21@140]
                  [10.9@10, 37.9@142]]
    :returns: np.dnarray with cartesian form

    """
    return convert_math_cart(ALPHA_math)


def convert_matrix_to_math(matrix):
    """
    inverse of convert_matrix_to_cart

    """
    return convert_cart_math(matrix)

def rmse(residual_vibration):
    '''
    Calculate the root mean square error for residual_vibration column matrix
    subtract each residual vibration from zero and taking the square root
    of the summation, rounding the result to the fourth decimal point
    Args:
        residual_vibration: numpy array
    Return: RMSE deviated from 0
    '''
    return round(np.sqrt(np.abs(residual_vibration) ** 2).mean(), 4)

def residual_vibration(ALPHA, W, A):
    '''
    Calculate the residual vibration between ALPHA matrix
    and solution W with intial vibration A
    Args:
        ALPHA : Influence coefficient matrix -> np.array
        A : Initial vibration column array -> np.array
        W : Solution balancing weight row vector -> np.array
    Return:
        residual_vibration column array -> np.array
    '''
    return ALPHA @ W + A

def ill_condition(alpha):
    # Checking ILL-CONDITIONED planes
    # Using the algorithm as in Darlow `Balancing of High Speed Machinery 1989 chapter 6`
    # Find the norm of the influence matrix columns
    alpha_value = alpha.copy()
    U = np.linalg.norm(alpha_value, axis = 0)
    # arrange the influence matrix by the magnitude of the norm (descending)
    index = np.argsort(U)[::-1]
    alpha_arranged_by_column_norm = alpha_value[:, index]
    def u(i):
        '''
        returns column vector by index i in alpha_arranged_by_column_norm
        '''
        return alpha_arranged_by_column_norm[:, i, np.newaxis]

    def normalized(vector):
        '''
        Normalize vector = vector / norm(vector)
        Arg:
            vector -> complex of vector
            returns: normalized vector
            '''

        return vector / np.linalg.norm(vector)

    sf =[]  # Significance Factor
    # find e = u/norm(u)
    e = normalized(u(0))
    sigma = (np.conjugate(e).T @ u(1)) * e
    for i in range(0, len(index)-1):
        # find v1 = u1 - (conjugate(e0).T * u2)e1
        v = u(i+1) - sigma
        # sf1 = norm(v1) / norm(u1)
        sf.append(np.linalg.norm(v) / np.linalg.norm(u(i+1)))
        # calculate e1 = v2 / norm(v2)
        e = normalized(v)
        sigma += (np.conjugate(e).T @ u(i+1)) * e
    ill_plane = []
    for i, factor in enumerate(sf):
        if factor <= 0.2:
            ill_plane.append(index[i+1])
    return ill_plane



import numpy as np
import sys
import yaml
import pytest
import test_tools
import hsbalance as hs


'''This module is for testing Least square model solver'''
np.random.seed(42)
@pytest.fixture()
def n():
    '''
    tests for various matrix sizes
    for n = 1000 it took around 5 minutes on my machine
    '''
    return 10

@pytest.fixture()
def m():
    '''
    tests for various matrix sizes
    for n = 1000 it took around 5 minutes on my machine
    '''
    return 30

@pytest.fixture()
def test_alpha(m, n):
    '''
    Creating alpha instance
    '''
    alpha = hs.Alpha()
    real = np.random.uniform(0, 1, [m, n])
    imag = np.random.uniform(0, 1, [m, n])
    alpha.add(real + imag * 1j)
    return alpha

@pytest.fixture()
def test_A(m):
    '''
    Creating alpha instance
    '''
    real = np.random.uniform(0, 10, [m, 1])
    imag = np.random.uniform(0, 10, [m, 1])
    return real + imag * 1j
# @pytest.fixture()
# def C(m):
#     C = np.ones((m,1))
#     return C

# @pytest.fixture
# def test_model_WLS(test_A, test_alpha, n, C):
#     '''
#     Creating a test model
#     '''
#     w = model.LeastSquares(test_A, test_alpha, C).solve(solver='WLS')
#     return w


def test_matrix_WLS(test_alpha, test_A, m):
    '''
    Testing the performance of model
    '''
    alpha = test_alpha.value
    print(alpha)
    C = np.ones((m,1)) # weight matrix coefficient C
    # Carry out the test for c = 0 in all vector C indcies
    for c0 in range(len(C)):
        C[c0] = 0
        w = hs.model.LeastSquares(test_A, test_alpha, C).solve('WLS')
        # Cross check with the equation w = - pseudo_iverse(diag(C).alpha).A
        expected = - np.linalg.pinv(np.diag(C.T[0]) @ alpha) @ test_A
        np.testing.assert_allclose(expected, w)
        C[c0] = 1



import numpy as np
import hsbalance.tools as tools
import warnings


class Alpha():

    """
    Docstring for ALPHA.
    Alpha is the an influence coefficient matrix
    Influence coefficient matrix is a representation of the change of vibration
    vector in a measuring point when putting a unit weight on a balancing plane.
    """

    def __init__(self:'Influence matrix', name:'string'=''):
        """
        Instantiate an instance of Alpha
        name: optional name of Alpha
        """
        self.name = name

    def add(self, direct_matrix:'np.array'=None, A:'intial_vibraion numpy.array'=None,
            B:'trial matrix numpy.array'=None, U:'trial weight row vector numpy.array'=None,
            keep_trial:'optional keep the previous trial weight in every succeeding trial'=False,
            name:'string'=''):
        '''
        Method to add new values for Alpha instance
        either the direct_matrix is needed or ALL of (A, B, U)
        Args:
            direct_matrix: numpy array M rows -> measuring points,
                        N columns -> balancing planes
            A: Initial vibration column array -> numpy array
            B: Trial matrix MxN array -> numpy array
            U: Trial weights row array -> numpy array
            alpha = (A - B) / U
        '''
        try:  # test if direct input
            _ = direct_matrix.shape # TODO raise error when matrix is 1 dim
            if direct_matrix.shape[0] >= direct_matrix.shape[1]:
                self.value = direct_matrix
            else:
                raise tools.CustomError('Number of rows(measuring points) should be '
                                  'equal or  more than the number of columns '
                                  '(balancing planes)!')
        except AttributeError:
            # if direct matrix is not input calculate it from A, B, U
            # test the exstiance of A, A0, B, U to calculate ALPHA
            try:
                all([A.shape, B.shape, U.shape])
                # Test dimensions
                if A.shape[1] > 1:
                    raise tools.CustomError('`A` should be column vector')
                elif U.ndim > 1:
                    raise tools.CustomError('`U` should be row vector')
                elif B.shape[0] != A.shape[0] or B.shape[1] != U.shape[0]:
                    raise tools.CustomError('`B` dimensions should match `A`and `U`')
                else:
                    if not keep_trial:
                        self.value = (B - A) / U
                    else:
                        _A_keep_trial = np.delete((np.insert(B, [0], A, axis=1)),
                                                 -1, axis=1)
                        self.value = (B - _A_keep_trial) / U
            except AttributeError:
                raise tools.CustomError('Either direct_matrix or (A,B,U) '
                                        'should be passed "numpy arrays"')
    def check(self, ill_condition_remove=False):
        '''
        Method to check the alpha value
            * check the symmetrical of the matrix (check for square matrix only,
            for square matrix it should be symmetric obeyin the reciprocity law)
            * check for ill conditioned planes:
                if for any reason two or more planes has independent readings
                for example [[1, 2 , 3], [2, 4, 6]] this is named as ill-conditioned planes
                as they does not carry new information from the system and considering them
                cause solution infliteration.
            ill_condition_remove = True : remove the ill_condition planes after the check
        '''
        self.M = self.value.shape[0]
        self.N = self.value.shape[1]
        if self.M == self.N:
            _check_sym = np.allclose(self.value, self.value.T, 0.1, 1e-06)
            if not _check_sym:
                warnings.warn('Warning: Influence Matrix is asymmetrical!')
                _check_status_sym = 'Influence Matrix is asymmetrical, check your data'
            else:
                _check_status_sym = 'Influence Matrix is symmetric --> OK'
        else:
            _check_status_sym = 'Not a square matrix --> no exact solution'

        # Checking ILL-CONDITIONED planes
        ill_plane = tools.ill_condition(self.value)
        if ill_plane:
            _check_ill_condition = 'Ill condition found in plane{}'.format(ill_plane)
            if ill_condition_remove:
                self.value = np.delete(self.value,[ill_plane], axis=1)
        else:
            _check_ill_condition ='No ill conditioned planes --> ok'
        return print('{}\n\n{}'.format(_check_status_sym, _check_ill_condition))



import numpy as np
import sys
import yaml
import pytest
sys.path.insert(0, '../')
from hsbalance import model
from hsbalance import tools
import test_tools
from hsbalance.CI_matrix import Alpha


tests, tests_id, timeout = test_tools.get_tests_from_yaml('splitting')


@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_split_LSE(param, expected):
    my_ALPHA = Alpha()
    A = tools.convert_matrix_to_cart(param[0]['A'])
    A0 = [0]
    try:
        direct_matrix = tools.convert_matrix_to_cart(param[0]['ALPHA'])
        my_ALPHA.add(direct_matrix=direct_matrix)
    except KeyError:
        B = tools.convert_matrix_to_cart(param[0]['B'])
        U = tools.convert_matrix_to_cart(param[0]['U'])
        my_ALPHA.add(A=A, B=B, U=U)
    try:
         A0 = tools.convert_matrix_to_cart(param[0]['A0'])
    except KeyError:
        pass
    expected_W = tools.convert_matrix_to_cart(expected)[0]
    my_model = model.LeastSquares(A-A0, my_ALPHA, name='simple_least_square')
    W = my_model.solve()
    # Splitting model for weight at plane 0
    angles = np.arange(100,300,5)  # angles
    split = my_model.create_split()
    split.split_setup(0, max_number_weights_per_hole=1, holes_available=[angles]
                                                   ,weights_available=[10])
    split.split_solve()
    W_split = tools.convert_math_cart(split.error(options='equ'))

    np.testing.assert_allclose(W_split, expected_W, rtol=0.05) # allowance 5% error
    # Splitting model for weight at plane 1
    expected_W = tools.convert_matrix_to_cart(expected)[1]
    angles = np.arange(-50,50,5)  # angles
    split = my_model.create_split()
    split.split_setup(1, max_number_weights_per_hole=1, holes_available=[angles]
                                                   ,weights_available=[20])
    split.split_solve()
    W_split = tools.convert_math_cart(split.error(options='equ'))

    np.testing.assert_allclose(W_split, expected_W, rtol=0.05) # allowance 5% error



import numpy as np
import sys
import yaml
import pytest
import test_tools
import warnings
import hsbalance as hs

'''This module is for testing ALPHA class'''
# Reading the test cases from config.yaml file
tests, tests_id, timeout = test_tools.get_tests_from_yaml('ALPHA_direct')
# Parametrized the test for all testcases
@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_ALPHA_direct(param, expected):
    '''
    Pytest function to test instantiate Alpha from direct_matrix
    Inputs:
        param, expected : from config.yaml file
    '''
    # Instantiate Alpha class
    my_ALPHA = hs.Alpha()
    # Add direct matrix to the instance
    my_ALPHA.add(direct_matrix = hs.convert_matrix_to_cart(param))
    expected = list(list(complex(x) for x in item) for item in expected)
    np.testing.assert_allclose(my_ALPHA.value, expected, rtol=0.05)  # allowance 5% error

tests, tests_id, timeout = test_tools.get_tests_from_yaml('ALPHA_from_matrices')

@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_ALPHA_from_matrices(param, expected):
    '''
    Function to test instantiating Alpha class from matrices
    param, expected : from config.yaml file
    '''
    for key, value in param[0].items():
        globals()[key] = value
    my_ALPHA = hs.Alpha()
    my_A = hs.convert_matrix_to_cart(A)
    my_B = hs.convert_matrix_to_cart(B)
    my_U = hs.convert_matrix_to_cart(U)
    print(A, B, U, keep_trial)
    my_ALPHA.add(A=my_A, B=my_B, U=my_U, keep_trial=keep_trial)
    expected = hs.convert_matrix_to_cart(expected)
    np.testing.assert_allclose(my_ALPHA.value, expected, rtol=0.05)  # allowance 5% error


# Test symmetric
@pytest.fixture()
def test_alpha():
    '''
    Creating alpha instance to test throwing faults
    '''
    return hs.Alpha()

def test_alpha_check(test_alpha):
    '''
    Test raising symmetric matrix warning
    '''
    with pytest.warns(UserWarning):
        test_alpha.add(direct_matrix=np.array([[1, 2], [2.8, 0.9]]))
        test_alpha.check()
# Test ill_condition
def test_alpha_ill(test_alpha):
    '''
    Test ill conditioned planes
    '''
    test_alpha.add(direct_matrix=np.array([[1, 3], [2.5, 6.5]]))
    with pytest.warns(UserWarning):
        test_alpha.check(ill_condition_remove=True)
        # ill_condition_remove should remove the second column as it is depending of the first
        # (multiplied by 2 approx.)
        np.testing.assert_allclose(test_alpha.value, np.array([[3], [6.5]]))

def test_direct_matrix_dim(test_alpha):
    '''
    Check raising error due to improper dimension of matrix
    '''
    with pytest.raises(hs.CustomError) as e_info:
        test_alpha.add(direct_matrix=np.array([[1, 2, 3], [4, 5, 6]]))
    assert 'Number of rows(measuring points)' in str(e_info)

def test_alpha_dim(test_alpha):
    '''
    Check raising error due to improper dimension of matrix
    '''
    with pytest.raises(hs.CustomError) as e_info:
        # wrong dim matrix
        test_alpha.add(direct_matrix=np.array([[1, 2, 3], [4, 5, 6]]))
    assert 'Number of rows(measuring points)' in str(e_info)
    with pytest.raises(hs.CustomError) as e_info:
        # not numpy array
        test_alpha.add(direct_matrix=[[2,3], [3,4]])
    assert 'numpy arrays' in str(e_info)
    with pytest.raises(hs.CustomError) as e_info:
        # missing U and B
        test_alpha.add(A=np.array([[1],[2]]))
    assert 'Either' in str(e_info)
    with pytest.raises(hs.CustomError) as e_info:
        # A mismatch dim
        test_alpha.add(A=np.array([[1,2],[1,2]]), B=np.array([[1,2], [2,3]]), U=np.array([1, 3]))
    assert '`A` should be column ' in str(e_info)
    with pytest.raises(hs.CustomError) as e_info:
        # B mismatch dim
        test_alpha.add(A=np.array([[1],[1]]), B=np.array([[1,2,3], [2,3,3]]), U=np.array([1, 3]))
    assert '`B` dimensions' in str(e_info)
    with pytest.raises(hs.CustomError) as e_info:
        # U mismatch dim
        test_alpha.add(A=np.array([[1],[1]]), B=np.array([[1,2], [2,3]]), U=np.array([[1, 3, 3, 4]]))
    assert '`U` should be row' in str(e_info)
    with pytest.raises(hs.CustomError) as e_info:
        # U mismatch dim
        test_alpha.add(A=np.array([[1],[1]]), B=np.array([[1,2], [2,3]]), U='this is non sense')
    assert 'numpy arrays' in str(e_info)



import time
from scipy.interpolate import make_interp_spline
import matplotlib.pyplot as plt
import scipy.optimize as opt
import numpy as np
import sys
import yaml
import pytest
import test_tools
import hsbalance as hs


def test_faults():
    with pytest.raises(hs.CustomError) as error:
        hs.LeastSquares([1, 2], [1, 2])

@pytest.fixture()
def test_alpha():
    '''
    Creating alpha instance to test throwing faults
    '''
    alpha = hs.Alpha()
    value = np.random.uniform(0, 10, [2, 2])
    alpha.add(value + value * 1j)
    return alpha

def test_A_dim():
    '''
    Test the dimension of A to be 2x1
    '''
    real = np.random.uniform(0, 10, [1, 2])
    imag = np.random.uniform(0, 10, [1, 2])
    A = real + imag * 1j
    with pytest.raises(hs.CustomError) as e_info:
        hs.LeastSquares(test_A_dim, test_alpha)
    real = np.random.uniform(0, 10, [2, 2])
    imag = np.random.uniform(0, 10, [2, 2])
    A = real + imag * 1j
    with pytest.raises(hs.CustomError) as e_info:
        hs.LeastSquares(test_A_dim, test_alpha)


@pytest.fixture()
def n():
    '''
    tests for various matrix sizes
    for n = 1000 it took around 5 minutes on my machine
    '''
    return 250

@pytest.fixture()
def test_big_alpha(n):
    '''
    Creating alpha instance to test throwing faults
    '''
    alpha = hs.Alpha()
    real = np.random.uniform(0, 10, [n, n])
    imag = np.random.uniform(0, 10, [n, n])
    alpha.add(real + imag * 1j)
    return alpha

@pytest.fixture()
def test_big_A(n):
    '''
    Creating alpha instance to test throwing faults
    '''
    real = np.random.uniform(0, 10, [n, 1])
    imag = np.random.uniform(0, 10, [n, 1])
    return real + imag * 1j

@pytest.fixture
def test_model_LSQ(test_big_A, test_big_alpha):
    '''
    Creating a test model
    '''
    w =  hs.LeastSquares(test_big_A, test_big_alpha).solve()
    return w


@pytest.mark.timeout(100)
def test_big_matrix_LSQ(test_big_alpha, test_big_A, test_model_LSQ):
    '''
    Testing the perfromance of model
    '''
    start = time.time()
    error = hs.residual_vibration(test_big_alpha.value, test_model_LSQ, test_big_A)
    end = time.time()
    print(error.shape)
    print ('Time elapsed = ', end-start)
    np.testing.assert_allclose(error, 0, atol=1e-5)

if __name__ == '__main__':
    def test_performance(n):
        alpha = hs.Alpha()
        real = np.random.uniform(0, 10, [n, n])
        imag = np.random.uniform(0, 10, [n, n])
        alpha.add(real + imag * 1j)
        real = np.random.uniform(0, 10, [n, 1])
        imag = np.random.uniform(0, 10, [n, 1])
        A= real + imag * 1j
        start = time.time()
        w =  hs.LeastSquares(A, alpha).solve()
        error = hs.residual_vibration(alpha.value, w, A)
        t = time.time() - start
        return round(t, 2)
    performance_time = []
    N = [2, 10, 50, 100, 200, 400, 600, 800]
    for n in N:
        performance_time.append(test_performance(n))
    print(N, performance_time)
    spline = make_interp_spline(N, performance_time)
    x = np.linspace(min(N), max(N), 500)
    y  = spline(x)
    plt.plot(x, y, label="Performace Test")
    plt.xlabel('N (dimension of a Squared Influence Coeffecient Matrix)')
    plt.ylabel('Time (seconds)')
    plt.title('Performance Test of LeastSquares model')
    plt.savefig('../data/performace_test.png')
    plt.show()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hsbalance",
    version="0.2.2",
    author="Maged M.Eltorkoman",
    author_email="newmaged@gmail.com",
    description="Python tools for Practical Modeling and Solving High Speed Rotor Unbalance Problem",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/MagedMohamedTurk/Turbomachinery-Rotors-Balancing',
    project_urls={
        "Bug Tracker": "https://github.com/pypa/sampleproject/issues",
    },
    install_requires=['cvxpy>=1.1.18', 'pandas>=1.3.5'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=[
        "Optimization",
        "Linear Programming",
        "Integer Programming",
        "Operations Research",
        "Rotor Balance",
        "Rotor Balancing",
        "Vibration Analysis",
        "Trim Balancing",
        "Rotor Dynamics",
        "Influence Coefficient Matrix Method",
        "Convex Optimization",
        "Field Engineering",
        "Python3",
        "CVXPY",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
)



import numpy as np
import sys
import yaml
import pytest
sys.path.insert(0, '../')
from hsbalance import tools

# Import test data from test_cases.yaml


# Testing functions against test_cases
def get_tests_from_yaml(function):
    """Collects test_cases data from config.yaml file

    :function: function name -> str
    :returns: tests -> list of tuples (input, expected)
                , tests_id -> list

    """
    try:
        with open('config.yaml') as f:
            test_cases = yaml.load(f, Loader=yaml.FullLoader)
            # Loading global parameters
            timeout = test_cases['timeout']  # TODO check time to be a number
    except (yaml.YAMLError, IOError):
        print('Error loading <config.yaml> file')
    tests = []
    tests_id = []
    for test_id, test in test_cases[function]['test_cases'].items():
        tests.append(test)
        tests_id.append(test_id)
    return tests, tests_id, timeout


# Testing convert_to_polar
tests, tests_id, timeout = get_tests_from_yaml('convert_to_polar')


@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_convert_to_polar(param, expected):
    output = tools.convert_to_polar(complex(param))
    error = max((x-y)/y for x, y in zip(output, expected))
    print('output=', output, '\n expected', expected,
          '\nerror', error)
    assert error < 0.01


# Testing convert_to_cartesian
tests, tests_id, timeout= get_tests_from_yaml('convert_to_cartesian')


@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_convert_to_cartesian(param, expected):
    output = tools.convert_to_cartesian(param)
    expected = complex(expected)
    error = abs(output - expected)
    print('output=', output, '\n expected', expected, '\nerror', error)
    assert error < 0.01


# Testing convert_math_cart
tests, tests_id, timeout  = get_tests_from_yaml('convert_math_cart')


@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_convert_math_cart(param, expected):
    output = tools.convert_math_cart(param)
    expected = complex(expected)
    error = abs(output - expected)
    print('output=', output, '\n expected', expected, '\nerror', error)
    assert error < 0.01


# Testing convert_matrix_to_cart
tests, tests_id, timeout = get_tests_from_yaml('convert_matrix_to_cart')


@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_convert_matrix_to_cart(param, expected):
    output = tools.convert_matrix_to_cart(param)
    try:
        expected = list(list(complex(x) for x in item) for item in expected)
    except ValueError:
        expected = list(complex(x) for x in expected)
    print(output, expected)
    expected = np.array(expected)
    print('output=', output, '\n expected', expected, '\nerror',
          (output - expected) / expected)
    # estimation rtol 5% error to compensate manual entry error
    np.testing.assert_allclose(output, expected, rtol=0.05)

# Testing rmse
tests, tests_id, timeout = get_tests_from_yaml('rmse')


@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_rmse(param, expected):
    output = tools.rmse(tools.convert_matrix_to_cart(param))
    error = output - expected
    print('output=', output, '\n expected', expected,
          '\nerror', error)
    assert error < 0.01

# Testing convert_cart_math
tests, tests_id, timeout = get_tests_from_yaml('convert_cart_math')


@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_convert_cart_math(param, expected):

    output = tools.convert_cart_math(complex(param))
    assert expected == str(output)



import cmath
import cvxpy as cp
import numpy as np
import hsbalance.tools as tools
from hsbalance.CI_matrix import Alpha
import pandas as pd
import warnings

class _Model:
    """Abstract class for models"""

    def __init__(self, A:'initial_vibration', alpha: 'instance of Alpha class', name:'string'=''):
        """ Instantiate the model
        Args:
        A: Initial vibration vector -> np.ndarray
        ALPHA: Influence coefficient matrix -> class Alpha
        name: optional name of the model -> string
        """
        self.name = name
        try:
            if A.shape[1] == 1:
                self.A = A
        except AttributeError:
            raise tools.CustomError('Either direct_matrix or (A,B,U) '
                                    'should be passed "numpy arrays"')
        except IndexError:
            raise tools.CustomError('`A` should be a column vector (Nx1) dimension')

        self.split_instance = []  # List of all related splits that has modified the solution
        # Test if Alpha is an instance of Alpha class
        if not isinstance(alpha, Alpha):
            raise tools.CustomError('Please create an `Alpha instance` first --> ex. alpha = Alpha()')
        else:
            self.ALPHA = alpha.value

        try:
            _ = self.ALPHA.shape  # Test that alpha.value returns np.array
            self.N = self.ALPHA.shape[1]  # Get N number of balancing_planes
            self.M = self.ALPHA.shape[0] # Get M number of measuring points
        except AttributeError:
            raise tools.CustomError('Missing valid ALPHA value')

    def solve(self):
        """
        Solve the problem and return W which is
        row vector of balancing weights
        """
        return self.W

    def expected_residual_vibration(self):
        """
        Returns the residual_vibration from tools module
        """
        return tools.residual_vibration(self.ALPHA, self.W, self.A)

    def rmse(self):
        """
        Returns the root mean squares from tools module
        """
        return tools.rmse(self.expected_residual_vibration())
    def create_split(self):
        """
        Factory method to create a split instance
        """
        return _Model.Split(self)

    class Split:
        '''
        Inner class for splitting weights
        '''
        def __init__(self, model_instance):
            """
            Enable accessing outer class variables
            args:
                model_instance: _Model class
            """
            self.model = model_instance

        def split_setup(self, balancing_plane_index, holes_available, weights_available, max_number_weights_per_hole=None,
                  max_weight_per_plane=None):
            """
            Split method to determine the optimized splitted weight
            over available holes and with some availabe weight types
            the optimization is mixed integer quadratic type that
            minimize the residule of the caluclated model weight from
            a combination of availabe weights on the availble holes
            Args:
                balancing_plane_index: the index of weight to be split
                    from the model. (model.W[index]) -> complex class
                holes_available: list of holes available, this can be
                    entered as numpy array (ex. np.arange(0, 100, 10) represents
                    the holes from 0 deg to 100 deg spaced 10 degress each)-> list
                weights_available: list of type of weights available
                    ex. [185, 90] represents that factory weights available
                    are 185 and 90 grams for instance. -> list
                max_number_weights_per_hole: maximum number of weights premissible
                    to be added in each hole. (ex. 1 -> only one weight is allowable
                    per hole) -> int
                max_weight_per_plane: max weight allowable per plane.
                    (ex 2000 -> maximum weight allowable per plane is 2 kg) -> float
            Returns: weight_available_matrix which represents all possible weights at
                    all possible holes.

            """
            self.balancing_plane_index = balancing_plane_index
            self.holes_available = holes_available
            self.weights_available = weights_available
            self.max_number_weights_per_hole = max_number_weights_per_hole
            self.max_weight_per_plane =  max_weight_per_plane
            self.weight = self.model.W[balancing_plane_index]


            # Vectorizing cmath `rect` function to be applied on holes
            vrect = np.vectorize(cmath.rect)
            # list to np.array
            if isinstance(holes_available, list):
                if holes_available:
                    holes_available = np.array(holes_available)
                else:
                    raise tools.CustomError('holes_available should contian at least one element')
            else:
                raise tools.CustomError('holes_available should be a list')
            if isinstance(weights_available, list):
                if weights_available:
                    weights_available = np.array(weights_available)
                else:
                    raise tools.CustomError('weights_available should contain at least one element')
            else:
                raise tools.CustomError('weights_available should be a list')
            # Transfer weights from row to column
            weights_available = weights_available[:, np.newaxis]
            # Make weight matrix in complex form
            weight_available_matrix = weights_available * vrect(1, holes_available * cmath.pi/180)  # from deg to rad
            self.weight_available_matrix = weight_available_matrix
            pass

        def split_solve(self):
            '''
            Mehtod to solve the splitting problem
            function takes the weight_available_matrix of the split instance
            and find the optimum number of weights types in the candidated holes_available
            that minimize the error.
            Args:
                split model instance
            Returns:
                S: numpy array of the same shape as weight_available_matrix of a model
            '''

            # define mixed integer matrix in cvxpy
            _S = cp.Variable(self.weight_available_matrix.shape, integer=True)
            # define objective function
            _Real = cp.sum(cp.multiply(np.real(self.weight_available_matrix), _S)) - np.real(self.weight)
            _Imag = cp.sum(cp.multiply(np.imag(self.weight_available_matrix), _S)) - np.imag(self.weight)
            _Residuals = cp.norm(cp.hstack([_Real, _Imag]))
            _obj_split = cp.Minimize(_Residuals)
            # constraints
            _const_splitting = [_S >= 0]
            if self.max_number_weights_per_hole:
                if isinstance(self.max_number_weights_per_hole, int):
                    _const_splitting += [cp.sum(_S, axis=0)
                                        <= self.max_number_weights_per_hole]
                else:
                    raise tools.CustomError('`max_number_weights_per_hole` should be integer number')
            if self.max_weight_per_plane:
                if isinstance(self.max_weight_per_plane, (float, int)):
                        _const_splitting += [cp.sum(_S.T @ self.weights_available, axis=0)
                                        <= self.max_weight_per_plane]
                else:
                    raise tools.CustomError('`max_weight_per_plane` should be float number')

            # solve
            _prob_S = cp.Problem(_obj_split, _const_splitting)
            _prob_S.solve(solver=cp.XPRESS)  # TODO make solver options PYTHON_MIP(), ECOS_BB
            _S = np.array(np.round(_S.value))
            self._S = _S
            self._prob_S = _prob_S.value
            return _S



        def results(self):
            '''
            Method to print comprehensive results of the solution
            transfer the numpy array to non-zeros dataframe
            Args:
                Instance of split model
            Return:
                df_splitting: panads dataframe
            '''
            df_splitting = pd.DataFrame(self._S, index=self.weights_available, columns=self.holes_available)
            df_splitting = df_splitting.loc[:, (df_splitting != 0).any(axis=0)]
            return df_splitting
        def error(self, options='error'):
            '''
            Method to determine different ways of error in the splitting.
            Args:
                Instance of split model
                options: kwarg to define the type of error.
                'error': default value returns the relative value of error.
                'problem_error': returns the accuracy of the solution.
                'equ': returns the equivilant weight of splitting solution
                    in math form.
            '''

            # Error equivilant weight after splitting
            _W_equ_split = np.sum(self.weight_available_matrix * self._S)
            self._W_equ_split = _W_equ_split
            _error = abs(self._W_equ_split - self.weight)
            # Estimate the error generated by using minmax instead of norm2 in the objective function
            _problem_error = (self._prob_S - _error) / abs(self.weight)
            if options == 'error':
                return _error
            elif options== 'problem_error':
                return _problem_error
            elif options == 'equ':
                return tools.convert_cart_math(_W_equ_split)
            else:
                raise tools.CustomError('Invalid options. Choose options="error" for solution error'
                                        'caused by splitting, options="problem_error" for solution'
                                        ' accuracy and options="equ" for equivilant weight.')
        def update(self, confirm=False):
            '''
            Method to override the model solution with the split solution for
            the balancing plane.
            confirm: Boolean= False: the default value the method does nothing.
                    It is important to inform the user that implementing this method
                    will cause a change of the optimum solution of the model.
            confirm = True: override the model solution, it is essential for user to
                            estimate the RMSE and residual_vibration change caused by splitting

            '''
            if confirm:
                self.model.W[self.balancing_plane_index] = self._W_equ_split
                self.model.split_instance.append(self)
            elif not confirm:
                warnings.warn('This will change your model optimum solution.'
                        'Choose confirm=True')
                self.model.W[self.balancing_plane_index] = self.weight
            return self.model.W


class LeastSquares(_Model):
    """subclass of Model
    solving the model using Least squares method, The objective function
    is to minimize the least squares of residual vibration.
    """


    def __init__(self, A:np.array, alpha:'instance of Alpha class', C=np.zeros(1), name=''):
        """ Instantiate the model
        Args:
        A: Initial vibration vector -> np.ndarray
        ALPHA: Influence coefficient matrix -> class Alpha
        C: Weighted Least squares coefficients
        name: optional name of the model -> string
        """
        super().__init__(A, alpha, name=name)
        if C.any():
            self.C = C
        else:
            self.C = np.ones(A.shape)
    def solve(self, solver='OLE'):
        '''
        Method to solve the model
        Args:
            solver:'OLE' Ordinary Least Squares method
            'Huber': Uses Huber smoother to down estimate the outliers.
        '''
        W = cp.Variable((self.N, 1), complex=True)
        if solver.upper() == 'OLE': # Ordinary least squares
            _objective = cp.Minimize(cp.sum_squares(self.ALPHA @ W + self.A))
        elif solver.upper() == 'HUBER':  # TODO test Huber solver for robust optimization
            _real = cp.real(self.ALPHA @ W + self.A)
            _imag = cp.imag(self.ALPHA @ W + self.A)
            _objective = cp.Minimize(cp.sum_squares(cp.huber(cp.hstack([_real, _imag]), M=0)))
        elif solver.upper() == 'WLS':  #  TODO test weighted least squares
            _objective = cp.Minimize(cp.sum_squares(cp.diag(self.C) @ (self.ALPHA @ W + self.A)))
        else:
            raise tools.CustomError('Unrecognized Solver name')
        prob = cp.Problem(_objective)
        prob.solve()
        self.W = W.value
        return W.value


class Min_max(_Model):
    """
    subclass of model: Solving the model using Minmax optimization method
    to minimize the maximum of residual_vibration.
    """

    def __init__(self, A:np.array, alpha:'instance of Alpha class', weight_const={}, name=''):
        """ Instantiate the model
        Args:
            A: Initial vibration vector -> np.ndarray
            alpha: instance of Alpha class
            weight_const: dict class of planes index and maximum permissible weight
        Returns: solution matrix W
        """
        self.weight_const = weight_const
        super().__init__(A, alpha, name=name)

    def solve(self, solver=None):
        '''
        Method to solve the Minmax model
        '''
        W = cp.Variable((self.N,1),complex=True)
        _objective = cp.Minimize(cp.norm((self.ALPHA @ W + self.A),"inf"))
        # Define weight constraints
        _constrains = []
        if self.weight_const != {}:
            try:
                for key, value in self.weight_const.items():
                    _constrains += [cp.norm(W[key]) <= value]
            except NameError:
                raise tools.CustomError('Invalid weight constraint format')
        else:
            pass
        prob=cp.Problem(_objective, _constrains)
        prob.solve()
        self.W = W.value
        return W.value


class LMI(_Model):
    """
    subclass of model
    solving using linear matrix inequality this is mainly to insert 'lazy constraints'
    Lazy constraints will relax the solution at certain planes (not to exceed certain vibration limit)
    """

    def __init__(self, A:np.array, alpha:'instance of Alpha class',
                 weight_const={},
                 critical_planes={}, V_max=None, name=''):
        """
        Args:
            A: Initial vibration vector -> np.ndarray
            alpha: Instance of Alpha class
            critical_planes:  set of critical planes
            weight_const: dict class of planes index and maximum permissible weight
            V_max: max vibration for non-critical_planes
        Return: Solution Matrix W
       """
        self.weight_const = weight_const
        self.critical_planes = list(critical_planes)
        self.V_max = V_max
        super().__init__(A, alpha, name=name)

    def solve(self, solver=None):
        '''
        solving the LMI model
        returns solution matrix W
        '''

        # Weight constraints
        N = self.N
        M = self.M
        _wc = np.zeros(N)
        if self.weight_const != {}:
            try:
                for key, value in self.weight_const.items():
                    _wc[key] = value
            except NameError:
                raise tools.CustomError('Invalid weight constraint format')
        else:
            pass

        # Identify critical planes
        if self.V_max:
            _Vm = self.V_max
        else:
            raise tools.CustomError('V_max is not specified')

        if self.critical_planes and len(self.critical_planes)>0:
            _list_cr = self.critical_planes
        else:
            raise tools.CustomError('Critical Planes are not set.')

        _ALPHA = self.ALPHA.copy()
        A = self.A.copy()
        _ALPHAcr = _ALPHA[_list_cr]
        Acr = A[_list_cr]
        _ALPHAncr = np.delete(_ALPHA, _list_cr, axis=0)
        _Ancr = np.delete(A, _list_cr, axis=0)
        # assign cvxpy variables
        _WR = cp.Variable((N, 1))
        _WI = cp.Variable((N, 1))
        _Vc = cp.Variable()

        _RRfcr = cp.diag(np.real(Acr) + np.real(_ALPHAcr) @ _WR-np.imag(_ALPHAcr) @ _WI)
        _IRfcr = cp.diag(np.imag(Acr) + np.imag(_ALPHAcr) @ _WR+np.real(_ALPHAcr) @ _WI)

        _RRfNcr = cp.diag(np.real(_Ancr) + np.real(_ALPHAncr) @ _WR - np.imag(_ALPHAncr) @ _WI)
        _IRfNcr = cp.diag(np.imag(_Ancr)+ np.imag(_ALPHAncr) @ _WR + np.real(_ALPHAncr) @ _WI)
        _zcr = np.zeros((_RRfcr.shape[0], _RRfcr.shape[1]))
        _Icr = np.eye(_RRfcr.shape[0])
        _zncr = np.zeros((_RRfNcr.shape[0], _RRfNcr.shape[1]))
        _Incr = np.eye(_RRfNcr.shape[0])

        _objective = cp.Minimize(_Vc)
        _LMI_real = cp.bmat( [
                                [_Vc*_Icr,        _RRfcr,    _zcr,        -_IRfcr],
                                [_RRfcr,          _Icr,      _IRfcr,          _zcr],
                                [_zcr,            _IRfcr,    _Vc*_Icr,        _RRfcr],
                                [-_IRfcr,         _zcr,       _RRfcr,          _Icr]
                                                                                    ])
        _LMI_imag =cp.bmat( [
                            [_Vm**2*_Incr,        _RRfNcr,    _zncr,        -_IRfNcr],
                            [_RRfNcr,          _Incr,      _IRfNcr,          _zncr],
                            [_zncr,            _IRfNcr,    _Vm**2*_Incr,        _RRfNcr],
                            [-_IRfNcr,         _zncr,       _RRfNcr,          _Incr]
                                                                                ])
        # Model weight constraints
        _const_LMI_w = []
        for i in range(N):
            _LMI_weight = cp.bmat( [
                                [_wc[i]**2,        _WR[i],    0,        -_WI[i]],
                                [_WR[i],          1,      _WI[i],          0],
                                [0,       _WI[i],    _wc[i]**2,        _WR[i]],
                                [-_WI[i],         0,       _WR[i],          1]
                                                                                          ])
            _const_LMI_w.append(_LMI_weight >> 0)
        _const_LMI_w.append( _LMI_real  >> 0)
        _const_LMI_w.append( _LMI_imag >> 0)
        # Stating the problem
        prob = cp.Problem(_objective, _const_LMI_w)
        prob.solve(cp.CVXOPT, kktsolver=cp.ROBUST_KKTSOLVER)
        W = _WR + _WI * 1j
        self.W = W.value
        return self.W



import numpy as np
import sys
import yaml
import pytest
import test_tools
import hsbalance as hs


'''This module is for testing Least square model solver'''
# Reading the test cases from config.yaml file, to add more tests follow the rules on the file
tests, tests_id, timeout = test_tools.get_tests_from_yaml('least_squares')


@pytest.mark.parametrize('param, expected',
                         tests,
                         ids=tests_id
                         )
@pytest.mark.timeout(timeout)
def test_WLS(param, expected):
    '''
    Testing instantiate LMI model and test it against test cases
    '''
    my_ALPHA = hs.Alpha()
    A = hs.convert_matrix_to_cart(param[0]['A'])
    A0 = [0]
    # It is acceptable to enter either direct_matrix or A,B,U matrices
    try:
        direct_matrix = hs.convert_matrix_to_cart(param[0]['ALPHA'])
        my_ALPHA.add(direct_matrix=direct_matrix)
    except KeyError:
        B = hs.convert_matrix_to_cart(param[0]['B'])
        U = hs.convert_matrix_to_cart(param[0]['U'])
        my_ALPHA.add(A=A, B=B, U=U)
    try:
         A0 = hs.convert_matrix_to_cart(param[0]['A0'])
    except KeyError:
        pass
    expected_W = hs.convert_matrix_to_cart(expected)
    my_model = hs.LeastSquares(A-A0, my_ALPHA, name='simple_least_square')
    W = my_model.solve(solver='WLS')
    print('Residual Vibration rmse calculated = ', my_model.rmse())
    print('Residual Vibration rmse from test_case = ',
          hs.rmse(hs.residual_vibration(my_ALPHA.value, expected_W, A)))
    print('expected_residual_vibration',
          hs.convert_matrix_to_math(my_model.expected_residual_vibration()))
    np.testing.assert_allclose(W, expected_W, rtol=0.05) # allowance 5% error

